// ----------------------------------------------------------------------------
// 16.07.99

#include <ctype.h>
#include "z11.h"
#include "FileRead.h"

// ----------------------------------------------------------------------------
void * FileReader::findBind(XStr name, Block <Binder> & dataBase) {
  int i=dataBase.index(Binder(name));
  if(i==-1) error("could not bind name '"+name+"' to any object.");
  return dataBase[i].userData;
}
// ----------------------------------------------------------------------------
bool String2Bool(XStr x, bool def) {
  const char * ptr = x;
  while(*ptr && !isalnum(*ptr)) ptr++;
  XStr s=ptr;
  s.downcase();
  if(s=="true" || s=="yes" || s=="1" || s=="y" || s=="t") return true;
  if(s=="false" || s=="no" || s=="0" || s=="n" || s=="f") return false;
  return def;
}
// ----------------------------------------------------------------------------
void FileReader::close() {
  while(!sections.isEmpty()) delete sections.pop();
  s.close();
}
// ----------------------------------------------------------------------------
void FileReader::rewindSection() {
  if(emptySection()) error("trying rewind() empty section");
  s.seekg(sections.top()->start);
}
// ----------------------------------------------------------------------------
void FileReader::error(XStr xmsg) {
  // check which line is it
  long o=s.tellg();
  s.clear();
  s.seekg(0);
  int line=1;
  int charno=0;
  while(s && o-->0)
    if(s.get()=='\n') { line++; charno=0; }
    else charno++;
  XStr msg = xprintf("%s (%d:%d) : '%s'",fileName.chars(),line,charno,xmsg.chars());
  if(!sections.isEmpty())
    msg+="\nSection stack: ";
  else msg+="\nSection stack was empty.";
  while(!sections.isEmpty()) { msg += sections.top()->name + ", "; delete sections.pop(); }

  EParseError e(msg);

# ifdef DSX
    throw e;
# else
    handleException(e);
# endif
}
// ----------------------------------------------------------------------------------
XStr FileReader::getString(XStr name, XStr def) {
  if(seekName(name)) def = getWord();
  return def;
}
// ----------------------------------------------------------------------------
// name is used as a tmp buffer
bool FileReader::getBool(XStr name, bool def) {
  if(seekName(name)) {
    name = getWord();
    def = String2Bool(name,def);
  }
  return def;
}
// ----------------------------------------------------------------------------------
// name is used as a tmp buffer
// unsetf enbles dec/hex/oct numbers reading
int FileReader::getInt(XStr name, int def) {
  if(seekName(name)) {
    name = getWord();
    istrstream s(name);
    s.unsetf(ios::basefield);
    s >> def;
  }
  return def;
}
// ----------------------------------------------------------------------------------
// name is used as a tmp buffer
float FileReader::getFloat(XStr name, float def) {
  if(seekName(name)) {
    name = getWord();
    istrstream(name) >> def;
  }
  return def;
}
// ----------------------------------------------------------------------------------
void FileReader::goOut() {
  delete sections.pop();
}
// ----------------------------------------------------------------------------------
void * FileReader::getBind(XStr name, XStr def, Block<Binder> & dataBase) {
  return findBind(getString(name,def),dataBase);
}
// ----------------------------------------------------------------------------------
bool FileReader::nextClassWord() {
  if(emptySection()) error("looking for class word in empty section "+sections.top()->name);
  XStr word;
  s.seekg(sections.top()->lastClassWord);
  word=getWord();
  while(word=="{") {
    int cnt=1;
    while(s && cnt) {
      word=getWord();
      if(word=="}") cnt--;
      if(word=="{") cnt++;
    }
    word=getWord();
  }
  if(word=="}") {
    if(sections.top()->start==0) error("unexpected }");
    else return false;
  }
  if(!s || word=="") return false;
  classWord=word;
  sections.top()->lastClassWord=s.tellg();
  return true;
}
// ----------------------------------------------------------------------------------
bool FileReader::open(XStr fname) {
  s.close();
  s.open(fname,ios::binary|ios::in);
  fileName=fname;
  s.tellg(); // iostream bug!
             // must to call this here, make it work properly next time
  if(s) {
    Section * sc = new Section();
    sc->name="wholeFile";
    sc->start=0;
    sc->lastClassWord=sc->start;
    sections.push(sc);
    return true;
  }
  else throw EFile("can't open "+fname);
}
// ----------------------------------------------------------------------------------
void FileReader::goIn() {
  XStr w=getWord();
  Section * sc = new Section();
  if(w=="{") sc->start=s.tellg();
  else {
    sc->name=w;
    sections.top()->lastClassWord=s.tellg(); // name shouldn't be reconized as a class word
    XStr w2=getWord();
    if(w2!="{") sc->start=-1;
    else sc->start=s.tellg();
  }
  sc->lastClassWord=sc->start;
  sections.push(sc);
}
// ----------------------------------------------------------------------------------
// skips comments
XStr FileReader::getWord() {
  XStr word;
  bool qstring=false, // ""
       qchar=false, // ''
       empty=true, // word is empty
       escape=false; // '\'
  //s >> ws;
  int c;
  while(s) {
    c=s.get();
    if(c==EOF) break;
    if(!qstring && !qchar) { // comment chacing only outside any quotes
      if(c=='/' && s.peek()=='/') {
        if(!empty) { s.putback(c); return word; } // not empty - return!
        else {
          while(s && c!='\n') c=s.get();
          continue;
        }
      }
      else if(c=='/' && s.peek()=='*') {
        if(!empty) { s.putback(c); return word; }
        else {
          while(s && !(c=='*' && s.peek()=='/')) c=s.get();
          s.get(); // gets '/'
          continue;
        }
      }
    }
    if(!escape && !empty) {
      if(!qstring && !qchar) {
        if(c=='\'' || c=='"') { s.putback(c); return word; }
      }
      else if( (qstring && c=='"') || (qchar && c=='\'') ) return word;
    }
    if(!escape && empty && (c=='"' || c=='\'')) { // start of string
      if(c=='"') qstring=true;
      if(c=='\'') qchar=true;
    }
    else if(!escape && c=='\\') escape=true;
    else if(!escape && !qstring && !qchar && isspace(c)) {
      escape=false;
      if(!empty) return word;
    }
    else { word+=c; empty=false; escape=false; }
  }
  return word; // in case of EOF
}
// ----------------------------------------------------------------------------------
bool FileReader::seekName(XStr name) {
  XStr word;
  name.downcase();
  rewindSection();
  while(s) {
    word=getWord();
    if(word=="}") return false;
    if(word=="{") {
      int cnt=1;
      while(s && cnt) {
        word=getWord();
        if(word=="}") cnt--;
        if(word=="{") cnt++;
      }
      if(!s) return false;
    }
    if(downcase(word)==name) return true;
  }
  return false;
}
// ----------------------------------------------------------------------------------




