// ----------------------------------------------------------------------------------
/* [z-11 project]
 * (c) 1999 Copyright Malekith
 * <malekith@topnet.pl> <http://malekith.topnet.pl>
 * For info about copying see file copying.z11.
 *
 * file "Terminal.cc", created on 06/07/99.
 * descr : terminal - logic stuff. gfx is in Console.cc
 * greetings to Quake authors.
 */
// ----------------------------------------------------------------------------------

#include <fstream.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>
#include <ctype.h>

#include "z11.h"
#include "WorldMap.h"
#include "Terminal.h"
#include "Input.h"
#include "EvKbd.h"

FILEINFO(Terminal);

// ----------------------------------------------------------------------------------
// globals
Term * term = new Terminal();
bool noRedir=false, noRedirStdErr=false;
// ----------------------------------------------------------------------------------
// externs
bool String2Bool(XStr x, bool def);
void fixWinSizes();

extern bool gameRun, quitFast, needRedraw;
extern char version[];
extern int windowHeight, windowWidth;
extern bool drawMap,compass;
extern float speed,ospeed;

// ----------------------------------------------------------------------------------
// utils
// ----------------------------------------------------------------------------------
// gets 'name [=] value' from src
void getNV(istream & src, XStr & n, XStr & v) {
  XStr buffer;
  readline(src,buffer);
  n=""; v="";
  const char * ptr = buffer;
  while(isspace(*ptr)) ptr++;
  for(;!isspace(*ptr) && *ptr!='='; ptr++) n+=*ptr;
  while(isspace(*ptr)) ptr++;
  if(*ptr=='=') ptr++;
  while(isspace(*ptr)) ptr++;
  if(*ptr) v=ptr;
}
// ----------------------------------------------------------------------------------
// this is for set table
void needRedrawIfFalse(void * data, void * newData) {
  *(bool*)data=*(bool*)newData; // always can assign
  if(*(bool*)data==false) needRedraw=true; // but sometimes should also redraw screen
}
// ----------------------------------------------------------------------------------
// core Terminal
Terminal::Terminal() 
  : historyIndex(0), logged(false), inMsgLoop(false), aliasCount(0)  { 
  if(!setuser(getlogin())) setuser("user");
  setVar v[] = {
    { "moveSpeed", svFloat, &speed, 0, "(0,10>" },
    { "turnSpeed", svFloat, &ospeed, 0, "(0,10>" },
    { "compass", svBool, &compass, needRedrawIfFalse, 0 },
    { "map", svBool, &drawMap, needRedrawIfFalse, 0 },
    { 0, svError, 0, 0, 0 }
  };
  vars = new setVar[sizeof(v)/sizeof(v[0])];
  memcpy(vars,v,sizeof(v)); // simply copy memory
}
// ----------------------------------------------------------------------------------
void Terminal::init() {
  con.init();
  ifstream s(txtData+"cmdlhist");
  int n;
  while(s)
    readline(s,history[n=history.add("")]); // isn't that nice ?
  historyIndex=n+1;
  print("z-11 Terminal\n(c) 1999 Malekith\n\n");
}
// ----------------------------------------------------------------------------------
Terminal::~Terminal() {
  delete [] vars;
  ofstream s(txtData+"cmdlhist",ios::out|ios::trunc);
  for(int i=history.size()-100>?0; i<history.size(); i++)
    s << history[i] << "\n";
}
// ----------------------------------------------------------------------------------
#define printf(args...) ({ sprintf(buffer, args); print(buffer); })
void Terminal::add(XStr line) {
  while(line.length()>con.w) {
    XStr d=line.at(0,con.w);
    line.at(0,con.w)="";
    con.add(d);
  }
  if(line.length()>0) con.add(line);
}
// ----------------------------------------------------------------------------------
void Terminal::messageLoop() {
  if(inMsgLoop) throw EDead("in msg loop twice - unhandled recurse");
  inMsgLoop=true;
  
  const XStr curs = "_";
  //clear_keybuf();
  kbd->eventWaitModeOn();
  con.login();
  add(prompt+curs);
  XStr cmdLine="";
  bool done=false;

  while(!done) {
    con.setLast(prompt+cmdLine+curs);
    kbd->waitForEvent();
    Event * ev;
    while( (ev=kbd->getEvent()) != 0 && !done ) {
      assert(ev->className()=="EvKbd");
      EvKbd * e = (EvKbd*)ev;
      if(e->isSpecial) {
        switch(e->data) {
          case EvKbd::up:
            if(historyIndex>0) historyIndex--;
            if(historyIndex<history.size() && historyIndex>=0)
               cmdLine=history[historyIndex];
            else cmdLine="";
            break;
          case EvKbd::down:
            if(historyIndex<history.size()) historyIndex++;
            if(historyIndex<history.size()) cmdLine=history[historyIndex];
            else cmdLine="";
            break;
	}
      }
      else switch(e->data) {
        case '~':
        case '\e':
          cmdLine="logout";
          con.setLast(prompt+cmdLine);
          done=true;
          break;
        case '\b':
          if(cmdLine.length()>0) cmdLine.at((int)cmdLine.length()-1,1)=""; break;
        case '\n':
        case '\r':
          con.setLast(prompt+cmdLine);
          if(executeB(cmdLine)) done=true;
	  if(!done && cmdLine!="")
            historyIndex=history.add(cmdLine)+1;
          cmdLine="";
          if(!done) add(prompt+curs);
          break;
        case 32 ... ('~'-1):
        case ('~'+1) ... 255:
          cmdLine+=XStr((char)e->data);
          break;
      }
      delete ev;
    }
  }
  con.logout();
  logoutB();
  kbd->eventWaitModeOff();
  inMsgLoop=false;
}
// ----------------------------------------------------------------------------------
bool Terminal::executeB(XStr c) {
  if(downcase(c)=="die") { quitFast=true; abort(); }
  // maybe all this stuff with streams will crash system -- it's emergency exit
  if(downcase(c)=="halt") halt(); // see above
  istrstream s(c);
  s.unsetf(ios::basefield); // to enable '0x', '0' prefixes for integers

  XStr p; s >> p; p.downcase();
  if(p=="") return false; // empty command

  if(p=="help" || p=="?") {
    p=""; s >> p;
    if(p=="") help("terminal");
    else help(p);
  }
  else if(p=="logout") return true;
  else if(p=="exit") { gameRun=false; return true; }
  else if(p=="font") {
    s >> p;
    try {
      con.loadFont(p);
      con.drawAll();
    } catch(EFile & e) {
      printf("font: "+e.format());
    }
  }
  else if(p=="shot") { // not impl
  }
  else if(p=="login") {
    loginB();
    s >> p;
    if(s && p.length()>0)
      if(!setuser(p))
        add("login: login invalid : "+p);
  }
  else if(p=="textcolor") {
    int c=0xffffff; float v=0;
    s >> c >> v;
    con.setTextColor(c,v);
  }
  else if(p=="window") {
    int x=0,y=0;
    s >> x >> y;
    if(x&&y) {
      windowHeight=y;
      windowWidth=x;
      fixWinSizes();
    }
  }
  else if(p=="add") {
    p=""; s >> p;
    ofstream f(p,ios::app|ios::out);
    if(!f) printf("add: cant open '%s'",p.chars());
    else {
      readline(s,p);
      f << p << endl;
      f.close();
    }
  }
  else if(p=="addline") {
    p=""; s >> p;
    XStr nn="";
    ifstream i(p,ios::in);
    if(i) { i.seekg(-1,ios::end); if(i.get()!='\n') nn="\n"; }
    ofstream f(p,ios::app|ios::out);
    if(!f) printf("addline: cant open '%s'",p.chars());
    else {
      readline(s,p);
      f << nn << p << endl;
      f.close();
    }
  }
  else if(p=="//" || p[0]=='#') {
    // do nothing - comment
  }
  else if(p=="exec") {
    p=""; s >> p;
    executeFile(p);
  }
  else if(p=="cat") {
    p=""; s >> p;
    ifstream s(p);
    printStream(s);
  }
  else if(p=="time") {
    time_t now; time(&now);
    tm t = *localtime(&now);
    printf("%d:%02d:%02d",t.tm_hour,t.tm_min,t.tm_sec);
  }
  else if(p=="date") {
    time_t now; time(&now);
    tm t = *localtime(&now);
    printf("%d/%02d/%02d",t.tm_year,t.tm_mon+1,t.tm_mday);
  }
  else if(p=="prompt") {
    readline(s,p);
    if(p.length()>0 && p.length()<10)
      prompt=p+" ";
    else setuser(user);
  }
  else if(p=="whoami") printf(user);
  else if(p=="set") setVariable(s);
  else if(p=="seemap")
    for(int i=0; i<map.w; i++) for(int j=0; j<map.h; j++) map.cell(i,j)->known=true;
  else if(p=="alias") {
    XStr n,v;
    getNV(s,n,v);
    if(n!="") {
      if(aliases.index(Alias(n))!=-1) {
        if(v=="") aliases[aliases.index(Alias(n))]=Alias(); // clear
        else {
          printf("alias: warning, redefinition of %s",n.chars());
          aliases[aliases.index(Alias(n))]=Alias(n,v);
        }
      }
      else {
        if(v=="") printf("alias: command must be given after alias name");
        int i;
        for(i=0; i<aliases.size(); i++)
          if(aliases[i].name=="") break; // try find empty
        if(i<aliases.size()) aliases[i]=Alias(n,v); // found
        else aliases.add(Alias(n,v)); // not found - add new
      }
    }
    else
      for(int i=0; i<aliases.size(); i++)
        if(aliases[i].name!="")
          printf("%s = '%s'",aliases[i].name.chars(),aliases[i].command.chars());
  }
  else if(p=="clear-aliases") aliases.reset();
  else if(p=="dump-aliases") {
    XStr fn; s >> fn;
    ofstream o(fn,ios::out|ios::trunc);
    o << "# z-11 aliases file\n"
         "# this file has been machine generated\n"
         "#\n";
    for(int i=0; i<aliases.size(); i++)
      o << "alias " << aliases[i].name << " = " << aliases[i].command << "\n";
    o.close();
  }
  else if(p=="html") {
    s >> p;
    try {
      XStr v=atURL(p);
      if(v=="") print("no contest found at url "+p);
      else print(formatHtml(v));
    } catch(EFile & e) {
      // for file not found error
      print("html: "+e.xmsg);
    }
  }
  else if(p=="clear") con.clear();
  else {
    int i=aliases.index(p);
    if(i==-1) add("unknown command: "+p);
    else {
      XStr a,parm; readline(s,parm);
      a=aliases[i].command;
      if(aliasCount>20)
        printf("infinite alias loop. trying run "+a);
      else {
        bool r=false;
        a.gsub(";;",(char)1); // could not use ASCII 1 on commands
        aliasCount++;
        while(a!="") {
          XStr b;
          if(a.contains(";")) {
            b=a.before(";");
            a=a.after(";");
          }
          else { b=a; a=""; }
          b.gsub((char)1,";");
          if(executeB(b+" "+parm)) r=true;
        }
        aliasCount--;
        return r;
      }
    }
  }
  return false;
}
// ----------------------------------------------------------------------------------
bool Terminal::setuser(XStr newuser) {
  if(newuser.length()<10) {
    user=newuser;
    prompt=user+" # ";
    return true;
  }
  else return false;
}
// ----------------------------------------------------------------------------------
void Terminal::executeFile(XStr name) {
  ifstream f(name);
  while(f) {
    XStr buffer;
    readline(f,buffer);
    if(executeB(buffer)) logoutB();
  }
  if(logged && !inMsgLoop) messageLoop();
}
// ----------------------------------------------------------------------------------
void Terminal::setVariable(istream & s) {
  XStr n,v;
  getNV(s,n,v);
  if(n=="") {
    XStr r="possible variables are [types] : ";
    for(setVar * p = vars; p->type!=svError; p++) {
      r+=XStr(p->name)+" [";
      switch(p->type) {
        case svInt: r+="int"; break;
        case svFloat: r+="float"; break;
        case svBool: r+="bool"; break;
        default: r+="!!! unknown type !!! this is bug !!!"; break;
        // won't halt() here ...
      }
      r+="], ";
    }
    r+="\nTry 'help <varible>' to find out more. It may don't work. Sorry.";
    printf(r);
    return;
  }
  n.downcase();
  setVar * p;
  for(p = vars; p->type!=svError; p++)
    if(n==downcase(XStr(p->name))) break;
  n="";
  switch(p->type) {
    case svError:
      printf("set: unknown veriable : '%s'",n.chars());
      break;
    case svBool:
      if(p->callBack) {
        bool b=String2Bool(v,true);
        p->callBack(p->data,&b);
      }
      else *(bool*)p->data=String2Bool(v,true);
      break;
    case svInt:
      {
        int x;
        istrstream st(v);
        st.unsetf(ios::basefield);
        st >> x;
        if(p->callBack) p->callBack(p->data,&x);
        else if(p->range) {
          if(isIn(x,p->range)) *(int*)p->data=x;
          else printf("set: value %d outside range %s",x,p->range);
        }
        else *(int*)p->data=x;
      }
      break;
    case svFloat:
      {
        float x;
        istrstream(v) >> x;
        if(p->callBack) p->callBack(p->data,&x);
        else if(p->range) {
          if(isIn(x,p->range)) *(float*)p->data=x;
          else printf("set: value %f outside range %s",x,p->range);
        }
        else *(float*)p->data=x;
      }
      break;
    default: halt();
  }
}
// ----------------------------------------------------------------------------------
bool Terminal::printLine(XStr line, int & n) {
  if(n++>=con.h-1) {
    con.add("-- [q],[esc] to quit, other for more --");
    kbd->waitForEvent();
    Event * e = kbd->getEvent();
    assert(e->className()=="EvKbd");
    int c = ((EvKbd*)e)->isSpecial ? 0 : ((EvKbd*)e)->data;
    if(c=='\e' || c=='q' || c=='Q') return true;
    n=1;
    con.setLast(line);
  }
  else con.add(line);
  return false;
}
// ----------------------------------------------------------------------------------
void Terminal::printStream(istream & s) {
  int n=0;
  XStr buffer;
  while(s) {
    buffer="";
    readline(s,buffer);
    if(!s && buffer=="") break;
    char lbuf[con.w+2];
    if(buffer!="") {
      const char * ptr=buffer.chars();
      while((int)strlen(ptr)>con.w) {
        const char * p2 = ptr+con.w-1;
        while(!isspace(*p2)&&p2>ptr) p2--;
        if(p2==ptr) p2=ptr+con.w-2;
        char * p3 = lbuf;
        while(ptr<=p2) *p3++=*ptr++;
        *p3=0;
        if(printLine(lbuf,n)) return;
      }
      if(printLine(ptr,n)) return;
    }
    else if(printLine("",n)) return;
  }
}
// ----------------------------------------------------------------------------------
void Terminal::print(XStr text) {
  istrstream s(text);
  printStream(s);
}
// ----------------------------------------------------------------------------------
bool isIt(XStr s, XStr tag) {
  return (downcase(s.at(0,tag.length()))==downcase(tag));
}
// ----------------------------------------------------------------------------------
XStr getTag(istream & s) {
  XStr r;
  bool tag=false;
  while(s) {
    int c=s.get();
    if(c==EOF) break;
    if(c=='<')
      if(r=="") {
        if(s.peek()=='!') {
          bool h=false;
          int c=s.get();
          while(s) {
            c=s.get();
            if(h && c=='>') break;
            if(c=='-' && s.peek()=='-') h=true;
            if(c!='-') h=false;
          }
        }
        else tag=true;
      }
      else { s.putback(c); return r; }
    if(isspace(c) && !tag && r!="") { s.putback(c); return r; }
    if(c=='>' && tag) { r+=c; return r; }
    if(!isspace(c)||tag) r+=c;
  }
  return r;
}
// ----------------------------------------------------------------------------------
XStr Terminal::atURL(XStr url) {
  XStr file, anchor;
  if(url.contains("#")) {
    file=url.before("#");
    anchor=url.after("#");
  }
  else file=url;
  ifstream s(file);
  if(!s) throw EFile("can't open html file "+file);
  while(s && !isIt(getTag(s),"<body")) { }
  if(anchor=="") {
    XStr r,t=getTag(s);
    while(s && !(isIt(t,"<a") && downcase(t).contains("name")))
      { t=getTag(s); r+=t+" "; }
    return r;
  }
  else {
    while(s) {
      XStr t=getTag(s);
      if(isIt(t,"<a")) {
        t.downcase(); // let's suppose that anchors are case insensitive ...
        XStr t2 = t.after("name");
        if(t2!="") {
          t2=t2.after("=");
          if(t2!="" && isspace(t2[0]))
            t2=t2.after(RXwhite);
          if(t2!="") {
            const char * ptr = t2;
            if(*ptr=='"') ptr++;
            XStr aname;
            while(*ptr && !XStr(">\" \t\n").contains(*ptr)) aname+=*ptr++;
            if(aname==downcase(anchor)) {
              // found!
              XStr r,t=getTag(s);
              while(s && !(isIt(t,"<a") && downcase(t).contains("name")))
                { t=getTag(s); r+=t+" "; }
              return r;
            }
          }
        }
      }
    }
  }
  return "";
}
// ----------------------------------------------------------------------------------
XStr Terminal::formatHtml(XStr what) {
  XStr pp; // preprocessed
  istrstream s(what);
  while(s) {
    XStr w=getTag(s);
    XStr x=downcase(w);

    w.gsub("&nbsp;"," ");
    w.gsub("&nbsp"," ");
    w.gsub("&lt;","<");
    w.gsub("&lt","<");
    w.gsub("&gt;",">");
    w.gsub("&gt",">");
    w.gsub("&amp;","&"); // don't allow '&amp' -> mistake may occur

    if(x=="") { } // do nothing
    else if(x=="<br>" && pp!="") pp+="\n";
    else if(x=="<p>" && pp!="") pp+="\n\n";
    else if(x[0]=='<') { } // for all other tags - do nothing
    else pp+=w+" ";
  }
  return pp;
}
// ----------------------------------------------------------------------------------
void Terminal::help(XStr topic) {
  XStr name=lngData+"manual.htm#";
  XStr h,head;
  if(h=="") h=atURL(name+topic);
  if(h=="") { h=atURL(name+"set_"+topic); head="SET variable:\n"; }
  if(h=="") { h=atURL(name+"alias_"+topic); head="Standard alias:\n"; }
  if(h!="") {
    h=formatHtml(h);
    h=head+h;
    print(h);
  }
  else print("not help found for topic "+topic);
}
// ----------------------------------------------------------------------------------
void Terminal::handleEvent(Event * ev) {
  if(ev->className()=="EvCommand") {
    execute( ((EvCommand*)ev) -> cmd );
    if(logged) messageLoop();
    delete ev;
  }
  else Term::handleEvent(ev);
}
// ----------------------------------------------------------------------------------



